@@ -4,6 +4,7 @@ import os
 from typing import List, Tuple, Callable, Callable, NoReturn, Union
 import threading
 import json
+import colorsys
 
 # Virtual Types
 # (h,w), np.bool
@@ -62,6 +63,12 @@ def vcat(*args) -> TColorImage:
         resize_as_needed(to_color(im)) for im in args
     ])
 
+def to_hsv(im: TColorImage) -> TColorImage:
+    if len(im.shape) == 3:
+        return cv2.cvtColor(im, cv2.COLOR_BGR2HSV)
+    else:
+        return to_hsv(to_color(im))
+
 def to_color(im: TAnyImage) -> TColorImage:
     '''
     Returns a color copy of the given color or gray image
@@ -90,6 +97,11 @@ def mask_to_image(mask: TMask) -> TGrayImage:
     '''
     return mask.astype(np.uint8) * 255
 
+def image_to_mask(image: TGrayImage) -> TMask:
+    mask = np.zeros(image.shape, dtype=np.bool)
+    mask[image > 127] = True
+    return mask
+
 def detect_color(
         color_img: TColorImage, 
         detection_color: TBGR8Color, 
@@ -118,6 +130,14 @@ def detect_color(
     # Return the mask
     return color_mask
 
+def detect_color_hsv(color_img, detection_low, detection_high):
+    hsv_img = to_hsv(color_img)
+    detection_low_arr = np.array(detection_low, dtype=np.uint8)
+    detection_high_arr = np.array(detection_high, dtype=np.uint8)
+    correct_color_mask = cv2.inRange(hsv_img, detection_low_arr, detection_high_arr)
+    return image_to_mask(cv2.blur(correct_color_mask, (7,7)))
+    
+
 def make_checkerboard(width, height, res = 100, fname = None):
     total_width = res * width
     total_height = res * height
@@ -157,6 +177,8 @@ class UIInfo(object):
             render_image = to_color(myui)
 
         # Make mask by finding discrete colors in sorted order
+        # Quantize to multiples of 10
+        mask_raw = mask_raw // 10
         mask_levels = sorted(set(mask_raw.ravel()))
         mask = np.zeros(mask_raw.shape, dtype=np.int)
         for (level_ix, each_level) in enumerate(mask_levels):
@@ -295,6 +317,15 @@ class UIManager(object):
             for each_callback in self.on_exit_callbacks:
                 each_callback(*args)
 
+    def get_known_pointers(self):
+        return self.pointers.keys()
+    
+    def get_position(self, pointer_name):
+        if pointer_name in self.pointers:
+            return self.pointers[pointer_name][0]
+        else:
+            return None
+
     def update_pointer(self, pointer_name, new_position):
         prev_button = None
         prev_position = None
@@ -340,49 +371,239 @@ class UIManager(object):
         self.call_all_on_hover(on_hover_args)
         self.call_all_on_exit(on_exit_args)
 
+def IOU(a, b):
+    return area(intersection(a,b)) / max(0.0001, area(union(a,b)))
+
+def area(bbox):
+    return bbox[2] * bbox[3]
+
+def union(a,b):
+    x = min(a[0], b[0])
+    y = min(a[1], b[1])
+    w = max(a[0]+a[2], b[0]+b[2]) - x
+    h = max(a[1]+a[3], b[1]+b[3]) - y
+    return (x, y, w, h)
+
+def intersection(a,b):
+    x = max(a[0], b[0])
+    y = max(a[1], b[1])
+    w = min(a[0]+a[2], b[0]+b[2]) - x
+    h = min(a[1]+a[3], b[1]+b[3]) - y
+    if w<0 or h<0:
+        return (0, 0, 0, 0)
+    else:
+        return (x, y, w, h)
+
+def is_central(bbox, w, h, margin = 0.05):
+    x, y, bw, bh = bbox
+    xmin_allowed = round(margin * w)
+    xmax_allowed = w - xmin_allowed
+    ymin_allowed = round(margin * h)
+    ymax_allowed = h - ymin_allowed
+    if (x >= xmax_allowed or x + bw <= xmin_allowed) or (y >= ymax_allowed or y + bh <= ymin_allowed):
+        return False
+    else:
+        return True
+
+def sqrDistance(pointA, pointB):
+    (Ax, Ay) = pointA
+    (Bx, By) = pointB
+    return (Ax - Bx) ** 2 + (Ay - By) ** 2
+
 class ColorPointer(object):
     def __init__(self, manager: UIManager,
-                 pointer_color: TBGR8Color = (54, 66, 106),
-                 threshold = 20):
+                 color_low = (0, 120, 30),
+                 color_high = (33, 255, 255),
+                 contour_threshold = 100):
         self.manager = manager
-        self.color = pointer_color
-        self.threshold = threshold
+        self.color_low = color_low
+        self.color_high = color_high
+        self.contour_threshold = contour_threshold
+        self.mask = None
+        self.mask_drawn = None
     
     def update(self, unwarped_frame):
+        found_fingers = False
+        finger_tips = []
         if unwarped_frame is not None:
             # We can do detection
-            color_mask = detect_color(unwarped_frame, self.color, self.threshold)
-            # Find the lowest y (top of image)
-            px = np.argwhere(color_mask)
-            if len(px) > 0:
-                # Just find the top pixel that's true
-                (y,x) = px[np.argmin(px, axis=0)[0]]
-                self.manager.update_pointer("finger", (x,y))
+            color_mask = detect_color_hsv(unwarped_frame, self.color_low, self.color_high)
+            self.mask = color_mask
+            #self.mask_drawn = to_color(mask_to_image(self.mask))
+            contours, contour_hierarchy = cv2.findContours(mask_to_image(self.mask), cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
+            self.mask_drawn = np.zeros(unwarped_frame.shape, dtype=np.uint8)
+            # Now let's filter to only big contours
+            good_contours = [c for c in contours if cv2.contourArea(c) > self.contour_threshold]
+            # Then, let's filter to only contours in the central area
+            (h,w) = color_mask.shape
+            contour_bboxes = [cv2.boundingRect(c) for c in good_contours]
+            central_contours = [good_contours[i] for i in range(len(good_contours)) if is_central(contour_bboxes[i], w, h)]
+            if len(central_contours) > 0:
+                hand_contour = max(central_contours, key=lambda ctr: cv2.contourArea(ctr))
             else:
-                self.manager.update_pointer("finger", None)
-        else:
-            self.manager.update_pointer("finger", None)
+                hand_contour = None
+
+            if hand_contour is not None:
+                hand_hull = cv2.convexHull(hand_contour, returnPoints=False)
+                # Find defects from the convex hull - these are the indents between
+                # fingers
+                #cv2.drawContours(self.mask_drawn, [hand_contour], -1, (0, 255, 0), 3)
+                defects = cv2.convexityDefects(hand_contour, hand_hull)
+                points_to_draw = []
+                # It should be at least 5% of the smaller screen dim
+                defect_distance_thres = min(h,w) // 20
+                n_lines = 0
+                finger_defects = []
+                for defect_ix, each_defect in enumerate(defects):
+                    start_point_ix, end_point_ix, farthest_point_ix, distance_to_farthest = each_defect.ravel()
+                    start_px = tuple(hand_contour[start_point_ix].ravel())
+                    middle_px = tuple(hand_contour[farthest_point_ix].ravel())
+                    end_px = tuple(hand_contour[end_point_ix].ravel())
+                    # Change from fixed to floating point
+                    distance_to_farthest = distance_to_farthest / 256
+                    # This has our fingers, perfectly, but also a bit of noise.
+                    # Let's filter by distance_to_farthest
+                    if distance_to_farthest > defect_distance_thres:
+                        finger_defects.append((start_px, middle_px, end_px, distance_to_farthest))
+                # In the most common case, the  hand is coming in from the bottom
+                # so let's sort by middle-X to get the convexity defect on the edges of each finger in order
+                finger_defects.sort(key=lambda defect: defect[1][0])
+                
+                # List comprehension to make a 10 color palette
+                known_colors = [
+                    (int(255 * col_b), int(255 * col_g), int(255 * col_r))
+                    for (col_r, col_g, col_b) in [
+                        colorsys.hsv_to_rgb(palette_hue, 0.9, 0.8)
+                        for palette_hue in np.linspace(0, 0.8, 10)
+                    ]
+                ]
+
+                # Now, we have the "defects", which specify start/end as the finger
+                # tips, and middle as the space between fingers. We want the opposite -
+                # start / end between fingers, and middle as the tip
+
+                fingers_merged = []
+                
+                # The width of the finger should be at most 20% of the screen
+                finger_merge_distance = (max(h,w) // 10) ** 2
+                # So, let's merge adjacent defects
+                for defect_ix in range(len(finger_defects)-1):
+                    L_start, L_middle, L_end, L_distance = finger_defects[defect_ix]
+                    R_start, R_middle, R_end, R_distance = finger_defects[defect_ix+1]
+                    # start/end are arbitrary here. Let's collect them both into small lists
+                    L_points = [L_start, L_end]
+                    R_points = [R_start, R_end]
+                    possible_adjacent_pairs = [(0,0), (0,1), (1,0), (1,1)]
+                    # Find the best one
+                    # pair_poss is an index into L_points, then R_points, which contains tuples
+                    # Using sqr distance
+                    best_pair = min(possible_adjacent_pairs, key=lambda pair_poss:sqrDistance(L_points[pair_poss[0]], R_points[pair_poss[1]]))
+                    
+                    # This gets us the best pairing! Use the closer points
+                    # as start/end respectively
+                    # We know these are length 2, so we can do 1 - INDEX to get "the other point"
+                    L_start, L_end = (L_points[1 - best_pair[0]], L_points[best_pair[0]])
+                    R_start, R_end = (R_points[best_pair[1]], R_points[1-best_pair[1]])
+
+                    if sqrDistance(L_end, R_start) < finger_merge_distance:
+                        # Just average L end with R start to find the shared point!
+                        new_start = L_middle
+                        # Hack - just take the bigger y
+                        new_middle = tuple([round((L_end[coord] + R_start[coord]) / 2) for coord in range(2)])
+                        new_end = R_middle
+                        fingers_merged.append((new_start, new_middle, new_end))
+
+                # Draw the detected fingers
+                for finger_ix, (start_px, middle_px, end_px) in enumerate(fingers_merged):
+                    # Drawing with color, because now they are stable-ish
+                    color = known_colors[min(len(known_colors), finger_ix)]
+                    cv2.line(self.mask_drawn, start_px, middle_px, color, 2)
+                    cv2.line(self.mask_drawn, middle_px, end_px, color, 2)
+
+                # Simplify for just finger tips
+                finger_tips = [middle_px for (start_px, middle_px, end_px) in fingers_merged]
+                found_fingers = True
+
+        if found_fingers:
+            # If we found fingers, then we assigned to finger tips.
+            # We'll call our fingers finger0...fingerN
+            # First let's get which ones we need to update,
+            # and their finger numbers
+            existing_pointers = [
+                (pointer_name, self.manager.get_position(pointer_name))
+                for pointer_name in self.manager.get_known_pointers()
+                if pointer_name.startswith('finger')
+            ]
+            
+            # Everything we've found this frame must be assigned / added
+            found_not_assigned = set(range(len(finger_tips)))
+            
+            # Anything we had previous must be deleted, if it wasn't assigned
+            prev_not_found = set([ptr_name for (ptr_name, ptr_loc) in existing_pointers])
+
+            for (ptr_name, prev_loc) in existing_pointers:
+                if len(found_not_assigned) == 0:
+                    # We've assigned everything we've got
+                    break
+                # Find the closest detected pointer
+                best_match = min(found_not_assigned, key=lambda finger_ix:sqrDistance(finger_tips[finger_ix], prev_loc))
+                # Move this pointer to the newly detected location
+                self.manager.update_pointer(ptr_name, finger_tips[best_match])
+                # And mark that we've assigned it
+                found_not_assigned.remove(best_match)
+                prev_not_found.remove(ptr_name)
+            
+            # We have 3 cases (at the start)
+            # len(found_not_assigned) > len(prev_not_found):
+            #   -- We assign every prev pointer then add some
+            #   -- So now, prev_not_found is empty
+            # len(prev_not_found) > len(found_not_assigned):
+            #   -- prev_not_found has stale pointers
+            #   -- Nothing left in found_not_assigned
+            # len(prev_not_found) == len(found_not_assigned):
+            #   -- They're both done!
+            # Thus, the order of add/remove doesn't really matter, as only
+            # one will happen
+
+            # Delete stale pointers
+            for ptr_name in prev_not_found:
+                self.manager.update_pointer(ptr_name, None)
+            
+            # Add newly found pointers
+            for finger_ix in found_not_assigned:
+                # Get a pointer name we haven't used yet
+                N = 0
+                next_pointer_name = f"finger{N}"
+                while next_pointer_name in self.manager.pointers:
+                    N += 1
+                    next_pointer_name = f"finger{N}"
+                self.manager.update_pointer(next_pointer_name, finger_tips[finger_ix])
+            
+            # Now we've updated all the finger pointers!    
+            
+
+            
 
 class UserSettings(object):
     def __init__(self, 
             webcam: Union[int, str], 
             hflip: bool, 
             vflip: bool,
-            skintone: TBGR8Color,
-            skintone_thresh: int):
+            skintone_low: TBGR8Color,
+            skintone_high: int):
         self.webcam = webcam
         self.hflip = hflip
         self.vflip = vflip
-        self.skintone = skintone
-        self.skintone_thresh = skintone_thresh
+        self.skintone_low = skintone_low
+        self.skintone_high = skintone_high
     def save(self, fpath: str):
         with open(fpath, 'w') as f:
             json.dump({
                 "webcam": self.webcam, 
                 "hflip": self.hflip, 
                 "vflip": self.vflip, 
-                "skintone": self.skintone, 
-                "skintone_thresh": self.skintone_thresh
+                "skintone_low": self.skintone_low, 
+                "skintone_high": self.skintone_high
             }, f, indent=4, sort_keys=True)
     @staticmethod
     def load(fpath: str):
@@ -395,13 +616,13 @@ class UserSettings(object):
                 "webcam", 
                 "hflip", 
                 "vflip", 
-                "skintone", 
-                "skintone_thresh"])
+                "skintone_low", 
+                "skintone_high"])
             ), "Config file is invalid"
         return UserSettings(**dct)
     @staticmethod
     def get_default():
-        return UserSettings(0, False, False, (44, 50, 85), 11)
+        return UserSettings(0, False, False, (0, 120, 30), (33, 255, 255))
     
     @staticmethod
     def load_global_settings():
@@ -430,7 +651,7 @@ class ARTable(object):
         self.ui = UIInfo(ui_name)
         self.detector = UIDetector(self.ui)
         self.manager = UIManager(self.ui)
-        self.finger = ColorPointer(self.manager, user_settings.skintone, user_settings.skintone_thresh)
+        self.finger = ColorPointer(self.manager, self.settings.skintone_low, self.settings.skintone_high)
         self.on_key = on_key
         self.is_running = False
         self.thread = None
@@ -449,21 +670,31 @@ class ARTable(object):
         additional_args = (cv2.CAP_DSHOW,) if not is_file else tuple([])
         cam = cv2.VideoCapture(self.webcam_id, *additional_args)
         has_pressed = False
+        is_paused = False
         while not self.stop_event.is_set():
-            ret, frame = cam.read(1)
-            if not ret:
-                break
-            # Apply flips as needed
-            frame = self.settings.transform_feed(frame)
+            if not is_paused:
+                ret, frame = cam.read(1)
+                if not ret:
+                    break
+                # Apply flips as needed
+                frame = self.settings.transform_feed(frame)
             self.detector.update_calibration(frame)
             unwarped = self.detector.unwarp_frame(frame)
             self.finger.update(unwarped)
             rendered = self.manager.draw()
-            cv2.imshow(self.window_name, hcat(unwarped if unwarped is not None else frame, rendered))
+            finger_mask = self.finger.mask_drawn#to_color(mask_to_image(self.finger.mask))
+            black = np.zeros_like(frame)
+            output_image = vcat(
+                hcat(frame, unwarped if unwarped is not None else black),
+                hcat(finger_mask if finger_mask is not None else black, rendered)
+            )
+            cv2.imshow(self.window_name, output_image)
             keys = cv2.waitKey(1) & 0xFF
-            if keys in map(ord,'qc'):
+            if keys in map(ord,'qcp'):
                 if not has_pressed:
                     has_pressed = True
+                    if keys == ord('p'):
+                        is_paused = not is_paused
                     if keys == ord('c'):
                         self.detector.toggle_calibration_allowed()
                     if keys == ord('q'):
